#include <ArduinoBLE.h>
#include <Arduino_LSM9DS1.h>
#include <TensorFlowLite.h>
#include <tensorflow/lite/experimental/micro/kernels/all_ops_resolver.h>
#include <tensorflow/lite/experimental/micro/micro_error_reporter.h>
#include <tensorflow/lite/experimental/micro/micro_interpreter.h>
#include <tensorflow/lite/schema/schema_generated.h>
#include <tensorflow/lite/version.h>
#include "model.h" // nasz model
const float gyroscopeThreshold = 300;
const int numSamples = 64;
int samplesRead = numSamples;
// zmienne globalne wykorzystywane przez TensorFlow Lite (TFLM)
tflite::MicroErrorReporter tflErrorReporter;
// pobieranie wszystkich opertorów TFLM; można zastąpić to
// pobieraniem tylko wybranych, by zmniejszyć skompilowanego programu
tflite::ops::micro::AllOpsResolver tflOpsResolver;
const tflite::Model* tflModel = nullptr;
tflite::MicroInterpreter* tflInterpreter = nullptr;
TfLiteTensor* tflInputTensor = nullptr;
TfLiteTensor* tflOutputTensor = nullptr;
// utworzenie statycznego bufora w pamięci dla TFLM
// jego wielkość jest zależna od modelu
constexpr int tensorArenaSize = 8 * 1024;
byte tensorArena[tensorArenaSize];
// macierz do mapowania gestów po indeksach
const char* GESTURES[] = {
  "Left arm",
  "Brake",
  "Hand front rotation",
  "Hand back rotation",
  "Hand left",
  "Hand right"
};
#define NUM_GESTURES (sizeof(GESTURES) / sizeof(GESTURES[0]))
int nbGesture = 0;
int oldNbGesture = 0;
int gestureTriggered;


void setup() {
  Serial.begin(115200);
  if (!IMU.begin()) {
    Serial.println("Failed to initialize IMU!");
    while (1);
  }
  // wypisanie prędkości próbek IMU
  Serial.print("Accelerometer sample rate = ");
  Serial.print(IMU.accelerationSampleRate());
  Serial.println(" Hz");
  Serial.print("Gyroscope sample rate = ");
  Serial.print(IMU.gyroscopeSampleRate());
  Serial.println(" Hz");
  Serial.println();
  // pobranie reprezentacji bitowej modelu TFLM
  tflModel = tflite::GetModel(model);
  if (tflModel->version() != TFLITE_SCHEMA_VERSION) {
    Serial.println("Model schema mismatch!");
    while (1);
  }
  // utowrzenie interoretera do uruchomienia modelu
  tflInterpreter = new tflite::MicroInterpreter(tflModel, tflOpsResolver, tensorArena, tensorArenaSize, &tflErrorReporter);
  // alokacja pamięci dla tensorów wejściowego i wyjściowego modelu
  tflInterpreter->AllocateTensors();
  // pobranie wskaźników na wejściowi i wyjściowy tensor modelu
  tflInputTensor = tflInterpreter->input(0);
  tflOutputTensor = tflInterpreter->output(0);

  pinMode(22, OUTPUT);
  pinMode(23, OUTPUT);
  pinMode(24, OUTPUT);
  digitalWrite(22, HIGH);
  digitalWrite(23, HIGH);
  digitalWrite(24, HIGH);

  // inicjalizacja transceivera BLE
  BLE.begin();
  Serial.println("BLE Central - LED control");

  // start skanowania peryferiów
  BLE.scan();
}

void loop() {
  BLEDevice peripheral = BLE.available();
  if (peripheral) {
    // odkryto urządzenie wypisz adres, nazwę lokalbą i serwer
    Serial.print("Found ");
    Serial.print(peripheral.address());
    Serial.print(" '");
    Serial.print(peripheral.localName());
    Serial.print("' ");
    Serial.print(peripheral.advertisedServiceUuid());
    Serial.println();
    if (peripheral.localName() != "MySmartglove") {
      return;
    }
    BLE.stopScan();
    // połącz z peryferiami 
    Serial.println("Connecting ...");

    if (peripheral.connect()) {
      Serial.println("Connected");
    } else {
      Serial.println("Failed to connect!");
      return;
    }

    // atrybuty odkrytych peryferiów
    Serial.println("Discovering attributes ...");
    if (peripheral.discoverAttributes()) {
      Serial.println("Attributes discovered");
    } else {
      Serial.println("Attribute discovery failed!");
      peripheral.disconnect();
      return;
    }

    // pobierz parametry LEDów 
    BLECharacteristic ledCharacteristic = peripheral.characteristic("e4297ee1-8c88-11ea-bc55-0242ac130003");

    if (!ledCharacteristic) {
      Serial.println("Peripheral does not have LED characteristic!");
      peripheral.disconnect();
      return;
    } else if (!ledCharacteristic.canWrite()) {
      Serial.println("Peripheral does not have a writable LED characteristic!");
      peripheral.disconnect();
      return;
    }

    while (peripheral.connected()) {
      float aX, aY, aZ, gX, gY, gZ;
      // czekaj na zauważalny ruch
      while (samplesRead == numSamples) {
        if (IMU.gyroscopeAvailable()) {
          IMU.readGyroscope(gX, gY, gZ);
          // suma wartości bezwzględnych
          float gSum = fabs(gX) + fabs(gY) + fabs(gZ);
          // sprawdź czy suma jest powyżej progu
          if (gSum >= gyroscopeThreshold) {
            // zresetuj licznik odebranych próbek
            samplesRead = 0;
            break;
          }
        }
      }
      // sprawdź, czy wszystkie wymagane próbki od czasu ostatniej aktywacji zostały odebrane
      oldNbGesture = nbGesture;
      while (samplesRead < numSamples) {
        // sprawdź, czy dostępne są nowe dane z akcelerometrów i żyroskopu
        if (IMU.accelerationAvailable() && IMU.gyroscopeAvailable()) {
          // odczytaj dane z akcelerometru i żyroskopu
          IMU.readAcceleration(aX, aY, aZ);
          IMU.readGyroscope(gX, gY, gZ);
          // normalizacja danych z IMU 0..1
          // tensor wejściowy
          tflInputTensor->data.f[samplesRead * 6 + 0] = (aX + 4.0) / 8.0;
          tflInputTensor->data.f[samplesRead * 6 + 1] = (aY + 4.0) / 8.0;
          tflInputTensor->data.f[samplesRead * 6 + 2] = (aZ + 4.0) / 8.0;
          tflInputTensor->data.f[samplesRead * 6 + 3] = (gX + 2000.0) / 4000.0;
          tflInputTensor->data.f[samplesRead * 6 + 4] = (gY + 2000.0) / 4000.0;
          tflInputTensor->data.f[samplesRead * 6 + 5] = (gZ + 2000.0) / 4000.0;
          samplesRead++;
          if (samplesRead == numSamples) {
            // inferencja
            TfLiteStatus invokeStatus = tflInterpreter->Invoke();
            if (invokeStatus != kTfLiteOk) {
              while (1);
              return;
            }
            // odczytaj nazwę gestu z tabeli dla modelu
            for (int i = 0; i < NUM_GESTURES; i++) {
              Serial.print(GESTURES[i]);
              Serial.print(": ");
              Serial.println(tflOutputTensor->data.f[i], 6);
              if ((tflOutputTensor->data.f[i]) > 0.6) {
                gestureTriggered = i;
                nbGesture++;
              }
            }
          }
        }
      }
      Serial.print("The gesture is :");
      Serial.println(GESTURES[gestureTriggered]);
      if (oldNbGesture != nbGesture) {
        if (gestureTriggered == 0) {
          ledCharacteristic.writeValue((byte)0x00);
          colors(0);
        }
        if (gestureTriggered == 2) {
          ledCharacteristic.writeValue((byte)0x02);
          colors(2);
        }
        if (gestureTriggered == 3) {
          ledCharacteristic.writeValue((byte)0x03);
          colors(3);
        }
        if (gestureTriggered == 4) {
          ledCharacteristic.writeValue((byte)0x04);
          colors(4);
        }
        if (gestureTriggered == 5) {
          ledCharacteristic.writeValue((byte)0x05);
          colors(5);
        }
      }
    }
    // peryferia odłączone, skanuj ponownie
    BLE.scan();
  }
}

int colors (int i) {
  if (i == 0) {
    for (int it1 = 0; it1 <= 1; it1++) {
      digitalWrite(22, LOW);
      digitalWrite(23, LOW);
      delay(500);
      digitalWrite(22, HIGH);
      digitalWrite(23, HIGH);
      delay(500);
    }
  }
  if (i == 1) {
    for (int it1 = 0; it1 <= 1; it1++) {
      digitalWrite(22, LOW);
      delay(500);
      digitalWrite(22, HIGH);
      delay(500);
    }
  }
  if (i == 2) {
    digitalWrite(23, LOW);
    delay(500);
    digitalWrite(23, HIGH);
    delay(500);
  }
  if (i == 3) {
    digitalWrite(22, LOW);
    delay(500);
    digitalWrite(22, HIGH);
  }
  if (i == 4) {
    digitalWrite(22, LOW);
    delay(500);
    digitalWrite(23, LOW);
    delay(500);
    digitalWrite(22, HIGH);
    digitalWrite(23, HIGH);
  }
  if (i == 5) {
    digitalWrite(23, LOW);
    delay(500);
    digitalWrite(22, LOW);
    delay(500);
    digitalWrite(22, HIGH);
    digitalWrite(23, HIGH);
  }
}
